﻿using System;
using System.Diagnostics;
using System.Threading;

namespace Synchronization
{
    class Program
    {

        static int thNumber = 80;
        static int countFor = 1250000;

        static int counter = 0;
        static void ThreadWithoutProtection()
        {
            for (int ii = 0; ii < countFor; ii++)
            {
                {
                    counter++;
                }
            }
        }

        static object locker = new object();
        static void ThreadWithLock()
        {
            for (int ii = 0; ii < countFor; ii++)
            {
                lock (locker)
                {
                    counter++;
                }
            }
        }



        static void ThreadWithMonitor()
        {
            for (int ii = 0; ii < countFor; ii++)
            {
                Monitor.Enter(locker);
                try
                {
                    counter++;
                }
                finally { Monitor.Exit(locker); }
            }
        }

        static void ThreadWithInterLock()
        {
            for (int ii = 0; ii < countFor; ii++)
            {
                Interlocked.Increment(ref counter);
            }
        }

        static Mutex mutex = new Mutex(false, "tu.kielce.pl mutex");
        static void ThreadWithMutex()
        {
            //it is very slow, there is 100 times less iterations
            for (int ii = 0; ii < countFor/100; ii++)
            {
                mutex.WaitOne();
                counter++;
                mutex.ReleaseMutex();
            }
        }

        static Semaphore sem = new Semaphore(1, 1);
        static void ThreadWithSemaphore()
        {
            //it is very slow, there is 100 times less iterations
            for (int ii = 0; ii < countFor/100; ii++)
            {
                sem.WaitOne();
                counter++;
                sem.Release();
            }
        }

        static SemaphoreSlim semSlim = new SemaphoreSlim(1, 1);
        static void ThreadWithSemaphoreSlim()
        {   //Notice. It isn't so slow as Semaphore
            for (int ii = 0; ii < countFor; ii++)
            {
                semSlim.Wait();
                counter++;
                semSlim.Release();
            }
        }


        static void ThreadWithLock2()
        {
            for (int ii = 0; ii < countFor/1000; ii++)
            {
                lock (locker)
                { 
                    for (int jj = 0; jj < 1000; jj++)
                        counter++;
                }
            }
        }

        static SpinLock spinLock = new SpinLock();
        static void ThreadWithSpinLock()
        {
            bool lockTaken = false;
            for (int ii = 0; ii < countFor; ii++)
            {
                lockTaken = false;
                spinLock.Enter(ref lockTaken);
                {
                    counter++;
                }
                if (lockTaken) spinLock.Exit(false);
            }
        }

        static void Main(string[] args)
        {

            //Thread.Sleep(0); // resign of the assigned time quantum
            //Thread.Sleep(1000); // sleep for 1000 ms
            //Thread.Sleep(TimeSpan.FromHours(1)); // sleep for 1 hour
            //Thread.Sleep(Timeout.Infinite); // sleep forever until break.


            Thread thread1 = new Thread(() =>
            {
                Console.WriteLine("Thread started and going to sleep for 3 s...");
                Thread.Sleep(3000);
                Console.WriteLine("Thread woke up and finished");
            });
            thread1.Start();
            Console.WriteLine("Main thread has started child thread and going to wait for it...");
            thread1.Join();
            Console.WriteLine("Main thread lived to see");

            //Unsafe thread increment
            Console.WriteLine("--------------- Unsafe thread increment -------------- ");
            Thread[] tharray = new Thread[thNumber];

            for (int ii = 0; ii < thNumber; ii++)
            {
                tharray[ii] = new Thread(ThreadWithoutProtection);
                tharray[ii].Name = "Thread " + ii.ToString();
            }
            Stopwatch sw = new Stopwatch();
            sw.Start();
            foreach (var th in tharray)
            {
                th.Start();
            }
            foreach (var th in tharray)
            {
                th.Join();
            }
            Console.WriteLine($"All unsafe threads finished in {sw.ElapsedMilliseconds} ms counter = {counter}");
            

            counter = 0;
            Console.WriteLine("--------------- lock thread increment -------------- ");
            Thread[] tharray2 = new Thread[thNumber];
            for (int ii = 0; ii < thNumber; ii++)
            {
                tharray2[ii] = new Thread(ThreadWithLock);
                tharray2[ii].Name = "Thread " + ii.ToString();
            }
            sw.Restart();
            foreach (var th in tharray2)
            {
                th.Start();
            }
            foreach (var th in tharray2)
            {
                th.Join();
            }
            Console.WriteLine($"All lock threads finished in {sw.ElapsedMilliseconds} ms counter = {counter}");


            counter = 0;
            Console.WriteLine("--------------- Monitor thread increment -------------- ");
            Thread[] tharray3 = new Thread[thNumber];
            for (int ii = 0; ii < thNumber; ii++)
            {
                tharray3[ii] = new Thread(ThreadWithMonitor);
                tharray3[ii].Name = "Thread " + ii.ToString();
            }
            sw.Restart();
            foreach (var th in tharray3)
            {
                th.Start();
            }
            foreach (var th in tharray3)
            {
                th.Join();
            }
            Console.WriteLine($"All Monitor threads finished in {sw.ElapsedMilliseconds} ms counter = {counter}");

            counter = 0;
            Console.WriteLine("--------------- Interlocked thread increment -------------- ");
            Thread[] tharray4 = new Thread[thNumber];
            for (int ii = 0; ii < thNumber; ii++)
            {
                tharray4[ii] = new Thread(ThreadWithInterLock);
                tharray4[ii].Name = "Thread " + ii.ToString();
            }
            sw.Restart();
            foreach (var th in tharray4)
            {
                th.Start();
            }
            foreach (var th in tharray4)
            {
                th.Join();
            }
            Console.WriteLine($"All Interlocked threads finished in {sw.ElapsedMilliseconds} ms counter = {counter}");


            counter = 0;
            Console.WriteLine("--------------- mutex thread increment -------------- ");
            Thread[] tharray5 = new Thread[thNumber];
            for (int ii = 0; ii < thNumber; ii++)
            {
                tharray5[ii] = new Thread(ThreadWithMutex);
                tharray5[ii].Name = "Thread " + ii.ToString();
            }
            sw.Restart();
            foreach (var th in tharray5)
            {
                th.Start();
            }
            foreach (var th in tharray5)
            {
                th.Join();
            }
            Console.WriteLine($"All mutex threads finished in {sw.ElapsedMilliseconds} ms counter = {counter}");


            counter = 0;
            Console.WriteLine("--------------- semaphore thread increment -------------- ");
            Thread[] tharray6 = new Thread[thNumber];
            for (int ii = 0; ii < thNumber; ii++)
            {
                tharray6[ii] = new Thread(ThreadWithSemaphore);
                tharray6[ii].Name = "Thread " + ii.ToString();
            }
            sw.Restart();
            foreach (var th in tharray6)
            {
                th.Start();
            }
            foreach (var th in tharray6)
            {
                th.Join();
            }
            Console.WriteLine($"All semaphore threads finished in {sw.ElapsedMilliseconds} ms counter = {counter}");

            counter = 0;
            Console.WriteLine("--------------- semaphore slim thread increment -------------- ");
            Thread[] tharray7 = new Thread[thNumber];
            for (int ii = 0; ii < thNumber; ii++)
            {
                tharray7[ii] = new Thread(ThreadWithSemaphoreSlim);
                tharray7[ii].Name = "Thread " + ii.ToString();
            }
            sw.Restart();
            foreach (var th in tharray7)
            {
                th.Start();
            }
            foreach (var th in tharray7)
            {
                th.Join();
            }
            Console.WriteLine($"All semaphore slim threads finished in {sw.ElapsedMilliseconds} ms counter = {counter}");


            counter = 0;
            Console.WriteLine("--------------- SpinLock thread increment -------------- ");
            Thread[] tharray8 = new Thread[thNumber];
            for (int ii = 0; ii < thNumber; ii++)
            {
                tharray8[ii] = new Thread(ThreadWithSpinLock);
                tharray8[ii].Name = "Thread " + ii.ToString();
            }
            sw.Restart();
            foreach (var th in tharray8)
            {
                th.Start();
            }
            foreach (var th in tharray8)
            {
                th.Join();
            }
            Console.WriteLine($"All SpinLock threads finished in {sw.ElapsedMilliseconds} ms counter = {counter}");


            counter = 0;
            Console.WriteLine("--------------- lock thread increment with biger granulity -------------- ");
            Thread[] tharray9 = new Thread[thNumber];
            for (int ii = 0; ii < thNumber; ii++)
            {
                tharray9[ii] = new Thread(ThreadWithLock2);
                tharray9[ii].Name = "Thread " + ii.ToString();
            }
            sw.Restart();
            foreach (var th in tharray9)
            {
                th.Start();
            }
            foreach (var th in tharray9)
            {
                th.Join();
            }
            Console.WriteLine($"All lock 2 threads finished in {sw.ElapsedMilliseconds} ms counter = {counter}");
        }
    }
}
